//---------------------------------------------------------------------------
#ifndef commonH
#define commonH
//---------------------------------------------------------------------------

//-----------------------------------------------------------------------------
#if defined(WIN32) || defined(_WIN32) || defined(WIN64) || defined(__BORLANDC__)
  #define WIN     1
  #define LINUX   0
#else
  #define WIN     0
  #define LINUX   1
  #define stricmp strcasecmp
  #define strupr(x) char* uptr = x; while ((*uptr = toupper(*uptr))) uptr++;
  #pragma GCC diagnostic ignored "-Warray-bounds"
#endif

//---------------------------------------------------------------------------
#define PROG_NAME               "CP/M Image Manager utility"
#define PROG_EXE_NAME           "cim"
#define PROG_VERSION            0x20
#define PROG_AUTHOR             "RM-TEAM, 3/2023"

//---------------------------------------------------------------------------
#define BYTE                    unsigned char
#define WORD                    unsigned short int
#define DWORD                   unsigned int
#define BOOL                    int
#define FALSE                   0
#define TRUE                    !FALSE

//---------------------------------------------------------------------------
#define BUFFER_SIZE             16384

#define FILL_BYTE               -27    // 0xE5 229
#define RECORD_SIZE             128
#define DIR_ENTRY_SIZE          32
#define EXTENT_SIZE             16384
#define RECORDS_PER_EXTENT      (EXTENT_SIZE / RECORD_SIZE)
#define DIR_ENTRIES_PER_RECORD  (RECORD_SIZE / DIR_ENTRY_SIZE)

#define MAX_CPM_FILE_SIZE       (8*1024*1024)       // 8MB
#define MAX_CPM_DISK_SIZE       (1024*1024*1024)    // 1GB

//-----------------------------------------------------------------------------
#define MAX_PATH_LEN            260
#define MAX_FILE_LEN            256
#define MAX_EXT_LEN             256
#define LINES_PER_DUMP_PAGE     20

//-----------------------------------------------------------------------------
#define ERR_OK                   0

#define ERR_CREATE_IMAGE        -1
#define ERR_OPEN_IMAGE          -2
#define ERR_IMAGE_SIZE          -3
#define ERR_READ_IMAGE          -4
#define ERR_WRITE_IMAGE         -5

#define ERR_CREATE_FILE         -6
#define ERR_OPEN_FILE           -7
#define ERR_READ_FILE           -8
#define ERR_WRITE_FILE          -9

#define ERR_DISK_FULL           -10
#define ERR_DIRECTORY_FULL      -11
#define ERR_FILE_EXIST          -12
#define ERR_FILE_NOT_EXIST      -13
#define ERR_BAD_FILE_NAME       -14
#define ERR_DIR_ENTRY           -15

#define ERR_IMG_EXT_MISSING     -16
#define ERR_CONF_FILE_MISSING   -17
#define ERR_IMG_DEF_MISSING     -18
#define ERR_IMG_DEF_ERROR       -19

#define ERR_IMG_SYS_TOO_LONG    -20
#define ERR_IMG_NO_SYS          -21
#define ERR_IMG_BAD_TYPE        -22

#define ERR_FILE_TOO_LONG       -23
#define ERR_PATH_NOT_ALLOWED    -24

#define ERR_BAD_OPTIONS         -94
#define ERR_UNKNOWN_OPTION      -95
#define ERR_NO_OPTION_VALUE     -96
#define ERR_BAD_OPTION_VALUE    -97

#define ERR_OUT_OF_MEMORY       -98
#define ERR_FATAL_ERROR         -99

#define ERR_PRINT_USAGE         -100

//-----------------------------------------------------------------------------
#define PAR_SPT                  1
#define PAR_XLT                  2
#define PAR_BLS                  4
#define PAR_EXM                  8
#define PAR_DSM                 16
#define PAR_DRM                 32
#define PAR_OFF                 64
#define PAR_OFFX               128
#define PAR_ALL                  (PAR_SPT | PAR_XLT | PAR_BLS | PAR_EXM | PAR_DSM | PAR_DRM | PAR_OFF)

//-----------------------------------------------------------------------------
#pragma pack(push, 1)

// file control block
typedef struct {
  char cUser;         // user number (0 to 15, 229 for empty directory entry)
  char caFN[8];       // file name
  char caFT[3];       // file type (ft[0].7 = Read Only, ft[1].7 = Hidden)
  BYTE btEX;          // extent (0 to 31)
  BYTE btS1;          // system reserved (always 0)
  BYTE btS2;          // high order bits of extent (only lower 7 bits)
  BYTE btRC;          // record count for extent ex (0 to 128)
  BYTE btaAU[16];     // allocation unit numbers that are occupied by file
} FCB;

#pragma pack(pop)

typedef struct {
  char cUser;
  char szFN[13];
  int iTotalRecs;
  int iSize;
  BOOL bReadOnly;
  BOOL bHidden;
} FILE_NAME_ATTR;

//---------------------------------------------------------------------------
typedef struct WorkParams {
  BOOL bCatalogue;          // show catalogue of image
  BOOL bFormatFull;         // perform full image format
  BOOL bFormatQuick;        // perform quick image format
  BOOL bAddFiles;           // add files tdo image
  BOOL bEraseFiles;         // erase files on image
  BOOL bExtractFiles;       // extract files from image
  BOOL bWriteSystem;        // write system to image
  BOOL bCopySystem;         // copy system from another image
  BOOL bExtractSystem;      // extract system from image
  BOOL bDumpHex;            // show file content as HEX-DUMP
  BOOL bDumpAscii;          // show file content as ASCII
  BOOL bOverwrite;          // overwrite existing files
  BOOL bImageTypeInfo;      // display info about type of Image
  BOOL bIgnoreInvalidDE;    // ignore invalid Directory entries

  BYTE *pbtBuff;            // pointer to working buffer
  FILE *fImage;             // pointer to image file structure
  FILE *fFile;              // pointer to working file structure
  int iFileLength;          // length of opened working file

  char szExePath[MAX_FILE_LEN]; // variables for file path/names
  char szExeName[MAX_FILE_LEN];
  char szImagePath[MAX_PATH_LEN];
  char szImageName[MAX_FILE_LEN];
  char szImageExt[MAX_EXT_LEN];
  char szFilePath[MAX_PATH_LEN];
  char szFileName[MAX_FILE_LEN];
  char szFileMask[MAX_FILE_LEN];
  char szFileNameCpm[MAX_FILE_LEN];

  FILE_NAME_ATTR fna;       // working file/name/attr structure
  BOOL bHasWildcard;        // flag that actual file name has wildcard
  char caCpmFileMask[11];   // working CP/M file name

  int iCpmFindPosition;     // position in directory on entry search
  char caCpmFindMask[11];   // search mask
  char cUsr;                // current target user area

  int *xlt;                 // pointer to translate table
  FCB *dir;                 // poiner to directory entry array
  BYTE *au;                 // pointer to allocation units bitmap array

  int iDirEntries;          // total directory entries
  int iAllocationUnits;     // total allocation units
  int iAllocationUnitSize;  // size of allocation unit
  int iAllocVectorSize;     // size of allocation vector
  int iRecordsPerAU;        // records per AU
  int iExtentsPerDirEntry;  // number of extents per dir entry
  int iAuPerDirEntry;       // AU per dir entry

  int iSectorsPerTrack;     // sectors per track
  int iSystemTracks;        // number of system tracks
  int iSystemSize;          // size of system
  int iTotalDiskSize;       // total disk size
  int iTrackSize;           // track size

} WORK_PARAMS;

//---------------------------------------------------------------------------
// opers.c
void catalogue(WORK_PARAMS *pWP);
void addFiles(WORK_PARAMS *pWP);
void eraseFiles(WORK_PARAMS *pWP, BOOL bForSingleAdd);
void extractFiles(WORK_PARAMS *pWP);
void asciiDump(WORK_PARAMS *pWP);
void hexDump(WORK_PARAMS *pWP);
void writeSys(WORK_PARAMS *pWP);
void copySys(WORK_PARAMS *pWP);
void extractSys(WORK_PARAMS *pWP);

// image.c
int readConfig(WORK_PARAMS *pWP);
int openImage(WORK_PARAMS *pWP);
void closeImage(WORK_PARAMS *pWP);
int writeBoot(WORK_PARAMS *pWP, BOOL bWriteTail);
int readSector(WORK_PARAMS *pWP, BYTE *pbtBuf, int iTrack, int iSector);
int writeSector(WORK_PARAMS *pWP, BYTE *pbtBuf, int iTrack, int iSector);
int readDirectory(WORK_PARAMS *pWP);
int writeDirectory(WORK_PARAMS *pWP);
int cleanDirectory(WORK_PARAMS *pWP);

// helper.c
void prepareNameAttr(WORK_PARAMS *pWP, int iDirEntryNum);
BOOL checkFileName(WORK_PARAMS *pWP, char *pszMask, const char *pszFN);
BOOL markAU(WORK_PARAMS *pWP, int iAUnum, BOOL bState);
int findFirstFreeAU(WORK_PARAMS *pWP);
int getFreeAuCount(WORK_PARAMS *pWP);
int getFreeDirEntriesCount(WORK_PARAMS *pWP);
int findFreeDirEntry(WORK_PARAMS *pWP);
int findDirEntry(WORK_PARAMS *pWP, char *pcaEnt, int iFrom);
int calcFileRecs(WORK_PARAMS *pWP, char *pcaEnt);
int cpmFindFirstFile(WORK_PARAMS *pWP, char *pcaCpmMask);
int cpmFindNextFile(WORK_PARAMS *pWP);
BOOL computeTrackSector(WORK_PARAMS *pWP, int iAU, int iRec, int *pT, int *pS);
int openCurrentFile(WORK_PARAMS *pWP, BOOL bCreate);
void closeCurrentFile(WORK_PARAMS *pWP);

// utils.c
void extractPathName(char *pszPath, char *pszName, char *pszExt, const char *pszSrc);
void pressEnter(const char *pszMsg);
int printError(int iErr, const void *pvPar1, const void *pvPar2);
int allocBuffer(void **ppvBuff, int iSize);
int parseNumber(const char *pszStr);
char* trimLeft(char *pszStr);
char* trimRight(char *pszStr);
char* trim(char *pszStr);

//---------------------------------------------------------------------------
#endif
//---------------------------------------------------------------------------

