//---------------------------------------------------------------------------
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include "common.h"

//---------------------------------------------------------------------------
int processArguments(int argc, char* argv[], WORK_PARAMS **ppWP);
void version(void);
void usage(void);

//---------------------------------------------------------------------------
int main(int argc, char* argv[])
{
  WORK_PARAMS *pWP = NULL;
  int iErr;

  do {
    version();

    if (argc == 1) {
      iErr = ERR_PRINT_USAGE;
      break;
    }

    // process of program arguments
    iErr = processArguments(argc, argv, &pWP);
    if (iErr != ERR_OK)
      break;

    // read configuration INI file
    iErr = readConfig(pWP);
    if (iErr != ERR_OK)
      break;

    // display only Image Type info?
    if (pWP->bImageTypeInfo)
      break;

    // open/create image
    iErr = openImage(pWP);
    if (iErr != ERR_OK)
      break;

    if (pWP->bFormatFull || pWP->bFormatQuick) {
      if (pWP->bFormatFull) {
        // create boot
        iErr = writeBoot(pWP, TRUE);
        if (iErr != ERR_OK)
          break;
      }

      // clean directory entries only
      iErr = cleanDirectory(pWP);
      if (iErr != ERR_OK)
        break;
    }

    // do operations with System tracks
    if (pWP->bWriteSystem) {
      writeSys(pWP);
      break;
    }
    if (pWP->bCopySystem) {
      copySys(pWP);
      break;
    }
    if (pWP->bExtractSystem) {
      extractSys(pWP);
      break;
    }

    // read directory
    iErr = readDirectory(pWP);
    if (iErr != ERR_OK)
      break;

    // do requested operation

    if (pWP->bCatalogue)  // print catalogue of image
      catalogue(pWP);
    else if (pWP->bAddFiles) {
      if (pWP->bOverwrite && !*pWP->szFileNameCpm)
        eraseFiles(pWP, FALSE);
      addFiles(pWP);
    }
    else if (pWP->bEraseFiles)
      eraseFiles(pWP, FALSE);
    else if (pWP->bExtractFiles)
      extractFiles(pWP);
    else if (pWP->bDumpAscii)
      asciiDump(pWP);
    else if (pWP->bDumpHex)
      hexDump(pWP);

  } while (FALSE);

  if (iErr == ERR_PRINT_USAGE) {
    usage();
    iErr = 0;
  }
  else if (iErr != ERR_OK)
    printf("\n");

  // * finally, free allocated memory *

  if (pWP) {
    // close image
    closeImage(pWP);
    // close working file
    if (pWP->fFile)
      fclose(pWP->fFile);
    // free buffers
    if (pWP->pbtBuff)
      free((void *) pWP->pbtBuff);
    if (pWP->xlt)
      free((void *) pWP->xlt);
    if (pWP->dir)
      free((void *) pWP->dir);
    if (pWP->au)
      free((void *) pWP->au);
    free(pWP);
  }

#ifdef _DEBUG
  getchar();
#endif

  return -iErr;
}
//---------------------------------------------------------------------------
int processArguments(int argc, char* argv[], WORK_PARAMS **ppWP)
{
  int iErr, i, iV;
  char *p, szStr[256];
  WORK_PARAMS *pWP = NULL;

  // processing and checking of input parameter
  do {
    // prepare working parameters structure
    *ppWP = NULL;
    iErr = allocBuffer((void *) &pWP, sizeof(WORK_PARAMS));
    if (iErr != ERR_OK)
      break;
    *ppWP = pWP;

    // alloc working buffer
    iErr = allocBuffer((void *) &pWP->pbtBuff, BUFFER_SIZE);
    if (iErr != ERR_OK)
      break;

    // default values
    pWP->cUsr = -1;

    // parse program path/name
    extractPathName(pWP->szExePath, pWP->szExeName, NULL, argv[0]);
    p = strrchr(pWP->szExeName, '.');
    if (p != NULL)
      *p = '\0';

    if (argc == 1) {
      iErr = ERR_PRINT_USAGE;
      break;
    }

    // parse image path/name
    extractPathName(pWP->szImagePath, pWP->szImageName, pWP->szImageExt, argv[1]);

    // process other parameters
    for (i = 2; i < argc; i++) {
      if (strcmp(argv[i], "-f") == 0) {
        pWP->bFormatFull = TRUE;
      }
      else if (strcmp(argv[i], "-q") == 0) {
        pWP->bFormatQuick = TRUE;
      }
      else if (strcmp(argv[i], "-u") == 0) {
        if (++i == argc) {
          iErr = printError(ERR_NO_OPTION_VALUE, argv[i - 1], NULL);
          break;
        }
        iV = parseNumber(argv[i]);
        if (iV < 0 || iV > 15) {
          iErr = printError(ERR_NO_OPTION_VALUE, argv[i - 1], argv[i]);
          break;
        }
        pWP->cUsr = (char) iV;
      }
      else if (strcmp(argv[i], "-c") == 0) {
        if (++i < argc) {
          extractPathName(pWP->szFilePath, pWP->szFileMask, NULL, argv[i]);
#if WIN
          if (*pWP->szFilePath != '\0')
#else
          if (strcmp(pWP->szFilePath, "./") != 0)
#endif
          {
            iErr = printError(ERR_PATH_NOT_ALLOWED, argv[i - 1], NULL);
            break;
          }
        }
        else
          strcpy(pWP->szFileMask, "*.*");
        pWP->bCatalogue = TRUE;
      }
      else if (strcmp(argv[i], "-a") == 0) {
        if (++i == argc) {
          iErr = printError(ERR_NO_OPTION_VALUE, argv[i - 1], NULL);
          break;
        }
        extractPathName(pWP->szFilePath, pWP->szFileMask, NULL, argv[i]);
        pWP->bAddFiles = TRUE;
      }
      else if (strcmp(argv[i], "-afn") == 0) {
        if (++i == argc) {
          iErr = printError(ERR_NO_OPTION_VALUE, argv[i - 1], NULL);
          break;
        }
        strcpy(pWP->szFileNameCpm, argv[i]);
      }
      else if (strcmp(argv[i], "-ow") == 0) {
        pWP->bOverwrite = TRUE;
      }
      else if (strcmp(argv[i], "-e") == 0) {
        if (++i == argc) {
          iErr = printError(ERR_NO_OPTION_VALUE, argv[i - 1], NULL);
          break;
        }
        extractPathName(pWP->szFilePath, pWP->szFileMask, NULL, argv[i]);
#if WIN
        if (*pWP->szFilePath != '\0')
#else
        if (strcmp(pWP->szFilePath, "./") != 0)
#endif
        {
          iErr = printError(ERR_PATH_NOT_ALLOWED, argv[i - 1], NULL);
          break;
        }
        pWP->bEraseFiles = TRUE;
      }
      else if (strcmp(argv[i], "-x") == 0) {
        if (++i == argc) {
          iErr = printError(ERR_NO_OPTION_VALUE, argv[i - 1], NULL);
          break;
        }
        extractPathName(pWP->szFilePath, pWP->szFileMask, NULL, argv[i]);
        pWP->bExtractFiles = TRUE;
      }
      else if (strcmp(argv[i], "-h") == 0) {
        if (++i == argc) {
          iErr = printError(ERR_NO_OPTION_VALUE, argv[i - 1], NULL);
          break;
        }
        extractPathName(pWP->szFilePath, pWP->szFileMask, NULL, argv[i]);
#if WIN
        if (*pWP->szFilePath != '\0')
#else
        if (strcmp(pWP->szFilePath, "./") != 0)
#endif
        {
          iErr = printError(ERR_PATH_NOT_ALLOWED, argv[i - 1], NULL);
          break;
        }
        pWP->bDumpHex = TRUE;
      }
      else if (strcmp(argv[i], "-t") == 0) {
        if (++i == argc) {
          iErr = printError(ERR_NO_OPTION_VALUE, argv[i - 1], NULL);
          break;
        }
        extractPathName(pWP->szFilePath, pWP->szFileMask, NULL, argv[i]);
#if WIN
        if (*pWP->szFilePath != '\0')
#else
        if (strcmp(pWP->szFilePath, "./") != 0)
#endif
        {
          iErr = printError(ERR_PATH_NOT_ALLOWED, argv[i - 1], NULL);
          break;
        }
        pWP->bDumpAscii = TRUE;
      }
      else if (strcmp(argv[i], "-w") == 0) {
        if (++i == argc) {
          iErr = printError(ERR_NO_OPTION_VALUE, argv[i - 1], NULL);
          break;
        }
        extractPathName(pWP->szFilePath, pWP->szFileMask, NULL, argv[i]);
        pWP->bWriteSystem = TRUE;
      }
      else if (strcmp(argv[i], "-i") == 0) {
        if (++i == argc) {
          iErr = printError(ERR_NO_OPTION_VALUE, argv[i - 1], NULL);
          break;
        }
        extractPathName(pWP->szFilePath, pWP->szFileMask, NULL, argv[i]);
        pWP->bCopySystem = TRUE;
      }
      else if (strcmp(argv[i], "-s") == 0) {
        if (++i == argc) {
          iErr = printError(ERR_NO_OPTION_VALUE, argv[i - 1], NULL);
          break;
        }
        extractPathName(pWP->szFilePath, pWP->szFileMask, NULL, argv[i]);
        pWP->bExtractSystem = TRUE;
      }
      else if (strcmp(argv[i], "-iti") == 0) {
        pWP->bImageTypeInfo = TRUE;
      }
      else if (strcmp(argv[i], "-iide") == 0) {
        pWP->bIgnoreInvalidDE = TRUE;
      }
      else {
        iErr = printError(ERR_UNKNOWN_OPTION, argv[i], NULL);
        break;
      }
    }
    if (iErr != ERR_OK)
      break;

    // check unallowed combinations of switches

    if (pWP->bFormatFull && pWP->bFormatQuick) {
      iErr = printError(ERR_BAD_OPTIONS, "-f", "-q");
      break;
    }

    if (!pWP->bAddFiles && pWP->bOverwrite) {
      iErr = printError(ERR_BAD_OPTIONS, "not allowed", "-ow");
      break;
    }

    if (!pWP->bAddFiles && *pWP->szFileNameCpm) {
      iErr = printError(ERR_BAD_OPTIONS, "not allowed", "-afn");
      break;
    }

    if ((pWP->bWriteSystem || pWP->bCopySystem || pWP->bExtractSystem || pWP->bImageTypeInfo) && pWP->cUsr != -1) {
      iErr = printError(ERR_BAD_OPTIONS, "not allowed", "-u");
      break;
    }

    if ((pWP->bWriteSystem || pWP->bCopySystem || pWP->bExtractSystem || pWP->bImageTypeInfo) && pWP->bIgnoreInvalidDE) {
      iErr = printError(ERR_BAD_OPTIONS, "not allowed", "-iide");
      break;
    }

    iV = 0;
    *szStr = '\0';
    if (pWP->bCatalogue) {
      iV++;
      strcat(szStr, "-c ");
    }
    if (pWP->bAddFiles) {
      iV++;
      strcat(szStr, "-a ");
    }
    if (pWP->bEraseFiles) {
      iV++;
      strcat(szStr, "-e ");
    }
    if (pWP->bExtractFiles) {
      iV++;
      strcat(szStr, "-x ");
    }
    if (pWP->bDumpHex) {
      iV++;
      strcat(szStr, "-h ");
    }
    if (pWP->bDumpAscii) {
      iV++;
      strcat(szStr, "-t ");
    }
    if (pWP->bWriteSystem) {
      iV++;
      strcat(szStr, "-w ");
    }
    if (pWP->bCopySystem) {
      iV++;
      strcat(szStr, "-i ");
    }
    if (pWP->bExtractSystem) {
      iV++;
      strcat(szStr, "-s ");
    }
    if (pWP->bImageTypeInfo) {
      iV++;
      strcat(szStr, "-iti ");
    }

    if (iV > 1) {
      iErr = printError(ERR_BAD_OPTIONS, szStr, "");
      break;
    }

    if (iV == 0 && !pWP->bFormatFull && !pWP->bFormatQuick) {
      iErr = ERR_PRINT_USAGE;
      break;
    }

    if (iV == 1 && !pWP->bAddFiles && (pWP->bFormatFull || pWP->bFormatQuick)) {
      iErr = printError(ERR_BAD_OPTIONS, "not allowed", "-f/-q");
      break;
    }

    if (*pWP->szFileNameCpm
        && (strchr(pWP->szFileMask, '?') != NULL || strchr(pWP->szFileMask, '*') != NULL)) {
      iErr = printError(ERR_BAD_OPTION_VALUE, "-a", pWP->szFileMask);
      break;
    }

  } while (FALSE);

  return iErr;
}
//---------------------------------------------------------------------------
void version(void)
{
  printf("\n%s v%d.%d\n%s\n\n",
         PROG_NAME, PROG_VERSION / 16, PROG_VERSION % 16, PROG_AUTHOR);
}
//-----------------------------------------------------------------------------
void usage(void)
{
#if WIN
  const char *szCpmName = "cpm_name";
  const char *szCpmNameWithPath = "[path/]cpm_name";
#else
  const char *szCpmName = "'cpm_name'";
  const char *szCpmNameWithPath = "'[path/]cpm_name'";
#endif

  printf("Usage:\n");
  printf("  %s imagefile.ext [ -u usr ] [ -iide ] -c [ %s ]\n", PROG_EXE_NAME, szCpmName);
  printf("  %s imagefile.ext [ -f | -q ] [ -a %s [ -afn %s ] [ -ow ] [ -u usr ] [ -iide ] ]\n", PROG_EXE_NAME, szCpmNameWithPath, szCpmName);
  printf("  %s imagefile.ext -e %s [ -u usr ] [ -iide ]\n", PROG_EXE_NAME, szCpmName);
  printf("  %s imagefile.ext -x %s [ -u usr ] [ -iide ]\n", PROG_EXE_NAME, szCpmNameWithPath);
  printf("  %s imagefile.ext -h %s [ -u usr ] [ -iide ]\n", PROG_EXE_NAME, szCpmName);
  printf("  %s imagefile.ext -t %s [ -u usr ] [ -iide ]\n", PROG_EXE_NAME, szCpmName);
  printf("  %s imagefile.ext -w [path/]in_system_file\n", PROG_EXE_NAME);
  printf("  %s imagefile.ext -i [path/]imagefile2.ext\n", PROG_EXE_NAME);
  printf("  %s imagefile.ext -s [path/]out_system_file\n", PROG_EXE_NAME);
  printf("  %s dummy.ext -iti\n", PROG_EXE_NAME);
  pressEnter(NULL);
  printf("    -c\tshow catalogue of the image\n");
  printf("    -f\tfull format of the image (image will be created if not exist)\n");
  printf("    -q\tquick format of the image (image will be created if not exist)\n");
  printf("    -a\tadd file(s) to the image (image will be created if not exist)\n");
  printf("    -afn  target 'cpm_name' without wildcards for single file\n");
  printf("\t  In this case file name in switch -a may be with long name,\n");
  printf("\t  but must not contain wildcards.\n");
  printf("    -ow\toverwrite existing file(s)\n");
  printf("    -e\terase file(s) from the image\n");
  printf("    -x\textract file(s) from the image\n");
  printf("    -h\tdisplay hex dump of the file\n");
  printf("    -t\tdisplay content of the ASCII file\n");
  printf("    -w\twrite system to the image from file\n");
  printf("    -i\twrite system to the image from another image of the same type\n");
  printf("    -s\textract system from the image and save to file\n");
  printf("    -iide\tignore invalid directory entries\n");
  printf("    -iti\tdisplay Image Type info\n");
  printf("    -u\tdetermines user number 0 to 15\n");
  printf("\tif not specified, files of all users are processed\n");
  printf("\tand for switch -a user 0 is always used\n");
  printf("    'imagefile.ext', 'imagefile2.ext' are image file names with extension\n");
  printf("    'cpm_name' is the file name in CP/M 8.3 format\n");
  printf("\t       which can contain wildcards '*' or '?'\n");
  printf("\t       and can be with dir-path of the file(s) to add or extract.\n");
#if LINUX
  printf("\t       NOTICE: 'cpm_name' with wildcards must be enclosed\n");
  printf("\t       in apostrophes to avoid wildcard filename expansion\n");
  printf("\t       by your filesystem.\n");
#endif
  printf("    'system_file' is binary file with CP/M system itself\n");
}
//---------------------------------------------------------------------------

